#include "config.h"

#include <gtk/gtk.h>

#include "boilerplate-log.h"
#include "boilerplate-application.h"
#include "boilerplate-window.h"
#include "boilerplate-preferences.h"

typedef struct
{
    BoilerplateWindow *window;
    GSettings         *settings;
} BoilerplateApplicationPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateApplication, boilerplate_application, GTK_TYPE_APPLICATION)

gboolean DEBUG = FALSE;
gboolean DEBUG_COLOR = TRUE;
gboolean VERSION = FALSE;
gboolean DISABLE_CSD = FALSE;

const GOptionEntry option_entries[] = {
    { "debug", 'd', 0, G_OPTION_ARG_NONE, &DEBUG, "Run in debug mode and show extra output", NULL },
    { "debug-no-color", '\0', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &DEBUG_COLOR, "Disable color when in debug mode", NULL },
    { "version", 'v', 0, G_OPTION_ARG_NONE, &VERSION, "Print release version and exit", NULL },
    { "disable-csd", 'c', 0, G_OPTION_ARG_NONE, &DISABLE_CSD, "Disable Client Side Decorations (CSD)", NULL },
    { NULL }
};

static GActionEntry action_entries[] = {
    { "preferences", boilerplate_application_action_preferences, NULL, NULL, NULL },
    { "shortcuts", boilerplate_application_action_shortcuts, NULL, NULL, NULL },
    { "help", boilerplate_application_action_help, NULL, NULL, NULL },
	{ "about", boilerplate_application_action_about, NULL, NULL, NULL },
	{ "quit", boilerplate_application_action_quit,  NULL, NULL, NULL },
};

static void
boilerplate_application_action_preferences (GSimpleAction *action,
                                            GVariant      *parameter,
                                            gpointer       user_data)
{
    GtkApplication *application;
    BoilerplateWindow *window;
    BoilerplatePreferences *preferences;

    application = GTK_APPLICATION (user_data);
    window = BOILERPLATE_WINDOW (gtk_application_get_active_window (application));

    preferences = boilerplate_preferences_new (window);

    gtk_window_present (GTK_WINDOW (preferences));
}

static void
boilerplate_application_action_shortcuts (GSimpleAction *action,
                                          GVariant      *parameter,
                                          gpointer       user_data)
{
    GtkApplication *application;
    BoilerplateWindow *window;
    GtkWidget *shortcuts_window = NULL;

    application = GTK_APPLICATION (user_data);
    window = BOILERPLATE_WINDOW (gtk_application_get_active_window (application));

    if (shortcuts_window == NULL)
    {
        GtkBuilder *builder;

        builder = gtk_builder_new_from_resource ("/io/github/julianrichen/boilerplate/ui/shortcuts.ui");
        shortcuts_window = GTK_WIDGET (gtk_builder_get_object (builder, "shortcuts"));

        g_signal_connect (shortcuts_window,
                          "destroy",
                          G_CALLBACK (gtk_widget_destroyed),
                          &shortcuts_window);

        g_object_unref (builder);
    }

    if (GTK_WINDOW (window) != gtk_window_get_transient_for (GTK_WINDOW (shortcuts_window)))
    {
        gtk_window_set_transient_for (GTK_WINDOW (shortcuts_window), GTK_WINDOW (window));
    }

    gtk_widget_show_all (shortcuts_window);
    gtk_window_present (GTK_WINDOW (shortcuts_window));
}

//TODO: Add Help content
static void
boilerplate_application_action_help (GSimpleAction *action,
                                     GVariant      *parameter,
                                     gpointer       user_data)
{
    GtkApplication *application;
    GtkWindow *window;
    GtkWidget *dialog;
    GError *error = NULL;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

    gtk_show_uri (window ? gtk_window_get_screen (GTK_WINDOW (window)) : gdk_screen_get_default (),
                  "help:boilerplate-for-c-gtk3",
                  gtk_get_current_event_time (),
                  &error);

    if (error)
    {
        dialog = gtk_message_dialog_new (window ? GTK_WINDOW (window) : NULL,
                                         GTK_DIALOG_MODAL,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_OK,
                                         "There was an error displaying help: \n%s",
                                         error->message);
        g_signal_connect (G_OBJECT (dialog), "response",
                          G_CALLBACK (gtk_widget_destroy),
                          NULL);

        gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
        gtk_widget_show (dialog);
        g_error_free (error);
    }
}

static void
boilerplate_application_action_about (GSimpleAction *action,
                                      GVariant      *parameter,
                                      gpointer       user_data)
{
    GtkApplication *application;
    GtkWindow *window;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

	static const char *authors[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const char *artists[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const gchar *translators =
        "Julian Richen <julian@richen.io>";

	gtk_show_about_dialog (window,
		                   "program-name", g_get_application_name (),
		                   "version", PACKAGE_VERSION,
		                   "copyright", "Copyright \xc2\xa9 "COPYRIGHT" Julian Richen",
		                   "comments", "Short description of application.",
		                   "authors", authors,
		                   "artists", artists,
		                   "translator-credits", translators,
		                   "website-label", "Website",
		                   "website", PACKAGE_URL,
		                   "logo-icon-name", gtk_window_get_default_icon_name (),
		                   "wrap-license", TRUE,
		                   "license-type", GTK_LICENSE_GPL_3_0,
		                   NULL);

}

static void
boilerplate_application_action_quit (GSimpleAction *action,
                                     GVariant      *parameter,
                                     gpointer       user_data)
{
    g_application_quit (G_APPLICATION (user_data));
}

static void
boilerplate_application_init_accelerators (GtkApplication *application)
{
	static const gchar *const accelmap[] = {
        "app.preferences", "<Primary>comma", NULL,
        "app.shortcuts", "<Primary>F1", NULL,
        "app.shortcuts", "<Primary><Shift>question", NULL,
        "app.help", "F1", NULL,
        "app.about", "F2", NULL,
        "app.quit", "<Primary>q", NULL,
		NULL
	};

	const gchar *const *it;

	for (it = accelmap; it[0]; it += g_strv_length ((gchar **)it) + 1)
    {
		gtk_application_set_accels_for_action (application, it[0], &it[1]);
	}
}

static void
boilerplate_application_startup (GApplication *application)
{
    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate *priv;
    GtkBuilder *builder;
    GMenuModel *app_menu;

    g_log (NULL, G_LOG_LEVEL_DEBUG, "Starting application");

    priv = boilerplate_application_get_instance_private (self);

    g_application_set_resource_base_path (G_APPLICATION (self), "/io/github/julianrichen/boilerplate");

	G_APPLICATION_CLASS (boilerplate_application_parent_class)->startup (G_APPLICATION (self));

    /* Set application name */
	g_set_application_name ("Boilerplate for C & GTK3");

    /* Set default icon */
    //TODO: Make icons better?: https://developer.gnome.org/ThemedIcons/
	gtk_window_set_default_icon_name ("io.github.julianrichen.boilerplate");

    /* Add app-menu */
    //TODO: Add ribbon menu as fallback
	g_action_map_add_action_entries (G_ACTION_MAP (application),
                                     action_entries, G_N_ELEMENTS (action_entries),
                                     application);

	boilerplate_application_init_accelerators (GTK_APPLICATION (application));

    builder = gtk_builder_new_from_resource ("/io/github/julianrichen/boilerplate/ui/app-menu.ui");
    app_menu = G_MENU_MODEL (gtk_builder_get_object (builder, "app-menu"));

    gtk_application_set_app_menu (GTK_APPLICATION (application), app_menu);

    g_object_unref (builder);
}

static void
boilerplate_application_activate (GApplication *application)
{
	// BoilerplateWindow *window;
    // window = boilerplate_window_new (BOILERPLATE_APPLICATION (application));
    // gtk_window_present (GTK_WINDOW (window));

    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate* priv = boilerplate_application_get_instance_private (self);

    //TODO: Investigate GList *windows
    if (priv->window == NULL)
        //TODO: Make CSD toggle-able, pass gpointer?
	    priv->window = boilerplate_window_new (BOILERPLATE_APPLICATION (application));

	gtk_window_present (GTK_WINDOW (priv->window));
}

static gint
handle_local_command_line (GApplication *self,
                           GVariantDict *options,
                           gpointer      user_data)
{

    if (VERSION)
    {
        g_print ("%s\n", PACKAGE_VERSION);
        return 0;
    }

    if (DISABLE_CSD)
    {
        //g_log (NULL, G_LOG_LEVEL_MESSAGE, "Client Side Decorations (CSD) have been disabled for this session.");
        g_print ("%s\n", "Client Side Decorations (CSD) have been disabled for this session.");
    }

    return -1;
}

static void
boilerplate_application_init (BoilerplateApplication *self)
{
    BoilerplateApplicationPrivate *priv;

    priv = boilerplate_application_get_instance_private (self);

	g_application_add_main_option_entries (G_APPLICATION (self), option_entries);

    /*
    priv->settings = g_settings_new ("io.github.julianrichen.boilerplate");
    */

    g_signal_connect (self, "handle-local-options", G_CALLBACK (handle_local_command_line), self);
}

static void
boilerplate_application_class_init (BoilerplateApplicationClass *class)
{
	G_APPLICATION_CLASS (class)->startup = boilerplate_application_startup;
	G_APPLICATION_CLASS (class)->activate = boilerplate_application_activate;
}

BoilerplateApplication *
boilerplate_application_new (void)
{
    g_log_set_handler (NULL,
                       G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
                       log_handler,
                       NULL);

    return g_object_new (BOILERPLATE_TYPE_APPLICATION,
                         "application-id", "io.github.julianrichen.boilerplate",
                         NULL);
}
